/* Copyright(C) 1997, Swedish Institute of Computer Science */

:- attribute fd_attribute/3.
%% index, domain mutable, suspensions mutable.

:- meta_predicate
	fd_global(:, ?, ?).

is_fd_variable(X) :-
	get_atts(X, fd_attribute(_,_,_)).

get_fd_domain_mutable(X, DomM) :-
	get_atts(X, fd_attribute(_,DomM,_)).

get_fd_domain(X, Dom) :-
	get_atts(X, fd_attribute(_,DomM,_)), !,
	get_mutable(Dom, DomM).
get_fd_domain(X, dom([[0|1]],0,1,2)) :-
	lk:is_lk_variable(X).

get_fd_suspensions(X, Lists) :-
	get_atts(X, fd_attribute(_,_,ListsM)), !,
	get_mutable(Lists, ListsM).
get_fd_suspensions(X, Lists) :-
	lk:is_lk_variable(X),
	Lists = '$fdlists'(0,0,[],[],[],[],[]).


solve(Constraint, DefPtr) :-
	check_arguments(Constraint, Domv),
	'$fd_get_indexicals'(DefPtr, 1, Constraint, _, _, Domv, Ixs),
	enqueue_first(Ixs, Q0, Q),
        evaluate(Q0, Q, G, G).

check_arguments(Constraint, Domv) :-
	'$fd_check_arguments'(Constraint, Domv), !.
check_arguments(Constraint, _) :-
	functor(Constraint, _, A),
	check_arguments_error(A, Constraint).

check_arguments_error(A, Constraint) :-
	arg(A, Constraint, Arg),
	(   var(Arg)
	;   integer(Arg),
	    \+prolog:'$large_data'(0, Arg, _)
	), !,
	B is A-1,
	check_arguments_error(B, Constraint).
check_arguments_error(A, Constraint) :-
	arg(A, Constraint, Arg),
	fd_argument_error(Constraint, A, Arg).

fd_argument_error(Constraint, A, Arg) :-
	'$fd_minint_maxint'(Minint, Maxint),
	fd_illarg(domain(integer,between(Minint,Maxint)), Constraint, A, Arg).

% avoid throwing domain variables
fd_illarg(Error, Constraint, ArgNo) :-
	prolog:'$term_variables'(Constraint, Vars),
	purify_vars(Vars),
	prolog:illarg(Error, Constraint, ArgNo).

fd_illarg(Error, Constraint, ArgNo, Culprit) :-
	prolog:'$term_variables'(Constraint-Culprit, Vars),
	purify_vars(Vars),
	prolog:illarg(Error, Constraint, ArgNo, Culprit).

% build queue, then evaluate
enqueue_first([]) --> [].
enqueue_first([Ix|Ixs]) -->
	{'$fd_mon_vars'(Ix, Vars)},
	enqueue_first(Vars, Ix),
	enqueue_first(Ixs).

enqueue_first([], Ix) --> [Ix].
enqueue_first([V|Vs], Ix) -->
	(   {var(V)} -> {'$fd_link'([val(V)], Vs-Ix)}
        ;   enqueue_first(Vs, Ix)
        ).


% evaluate(HeadOfInds, TailOfInds, HeadOfGlobals, TailOfGlobals) :- !.
% 
% Assume Why is a sum of the following bits:
% 0x10	if became ground
% 0x8	if became ground and there are such suspensions
% 0x4	if max pruned and there are suspensions for that
% 0x2	if min pruned and there are suspensions for that
% 0x1	if dom pruned and there are suspensions for that
% 
evaluate(Q0, _, GQ0, GQ) :- var(Q0), !, evaluate_global(GQ0, GQ).
evaluate([Ix|Ixs], IxTail, GQ, GQTail) :-
	'$fd_evaluate_indexical'(Ix, Ixs1, Ixs, IxTail, IxTail1, GQ1, GQ, GQTail, GQTail1),
	evaluate(Ixs1, IxTail1, GQ1, GQTail1).

evaluate_global(GQ0, _) :- var(GQ0), !.		% Both queues empty ==> stop.
evaluate_global([Global|GQ], GQTail) :-
	Global = global(StateM,Constraint,_Status,Ent,_),
	get_mutable(State, StateM),
	dispatch_prune_and_enqueue(Constraint, State, NewState,
				   Ent, Q0, Q, GQTail, GQTail1),
	'$fd_global_told'(Global),
	update_mutable(NewState, StateM),
	evaluate(Q0, Q, GQ, GQTail1).

prune_and_propagate(Pruned, Set) :-
	'$fd_in_set'(Pruned, Set, Q0, Q, GQ0, GQ),
	evaluate(Q0, Q, GQ0, GQ).

propagate_interval(Pruned, Min, Max) :-
	'$fd_in_interval'(Pruned, Min, Max, Q0, Q, GQ0, GQ),
	evaluate(Q0, Q, GQ0, GQ).

propagate_value(ListsM) :-
	'$fd_enqueue_all'(ListsM, Q0, Q1, Q1, Q, GQ0, GQ1, GQ1, GQ),
	evaluate(Q0, Q, GQ0, GQ).

%%% New global propagator API.

fd_global_tell(Var, FDSet) -->
	{'$fd_tell'(Var, FDSet, RC)},
	fd_global_told(RC, Var).

fd_global_told(0, _) --> [].
fd_global_told(1, Var) --> [dom(Var)].
fd_global_told(3, Var) --> [min(Var)].
fd_global_told(5, Var) --> [max(Var)].
fd_global_told(7, Var) --> [minmax(Var)].

verify_attributes(Var, Other, Goals) :-
	'$get_attributes'(Var, Atts, 1),
	verify_attributes1(Other, Atts, Goals).

verify_attributes1(Other, Atts, Goals) :-
	integer(Other),
	(   prolog:'$large_data'(0, Other, _) ->
	    fd_argument_error(_=Other, 2, Other)
	;   true
	),
	Atts = v(_,_,DomM,ListsM),
	get_mutable(Lists, ListsM),
	'$fd_set_singleton'(Other, DomM), % update_mutable(Other..Other, DomM),
	(   arg(2, Lists, 0) -> Goals = []
	;   Goals = [propagate_value(ListsM)]
	).
verify_attributes1(Other, Atts, Goals) :-
	var(Other),
	'$get_attributes'(Other, _, 1), !,
	'$put_attributes'(Fresh, Atts),
	Goals = [clpfd:'x=y'(Fresh,Other)].
verify_attributes1(Other, Atts, []) :-
	var(Other),
	'$put_attributes'(Other, Atts).


/* This caused problems as the C code assumed fixed offsets between
   domain mutables and v/4 structures.

verify_attributes(Var, Other, Goals) :-
	get_atts(Var, fd_attribute(_,DomM,ListsM)), % now guaranteed to succeed
	verify_attributes(Other, DomM, ListsM, Goals).

verify_attributes(Other, DomM, ListsM, Goals) :-
	integer(Other),
	get_mutable(Lists, ListsM),
	'$fd_set_singleton'(Other, DomM), % update_mutable(Other..Other, DomM),
	(   arg(2, Lists, 0) -> Goals = []
	;   Goals = [propagate_value(ListsM)]
	).
verify_attributes(Other, DomM, ListsM, Goals) :-
	var(Other),
	get_atts(Other, fd_attribute(_,_,_)), !,
	put_atts(Fresh, fd_attribute(0,DomM,ListsM)),
	Goals = [clpfd:'x=y'(Fresh,Other)].
verify_attributes(Other, DomM, ListsM, []) :-
	var(Other),
	put_atts(Other, fd_attribute(0,DomM,ListsM)).
*/

/* This is a more ambitious attempt to unify domain variables.
   It involves merging the suspension lists and intersecting the domains.
   However, this method breaks down now that we cache the domain mutables
   in the indexicals terms.

verify_attributes(Other, DomM, ListsM, Goals) :-
	var(Other),
	get_atts(Other, fd_attribute(_,OtherDomM,OtherListsM)), !,
	get_mutable(Dom, DomM),
	get_mutable(Lists, ListsM),
	get_mutable(OtherDom, OtherDomM),
	get_mutable(OtherLists, OtherListsM),
	merge_lists(Lists, OtherLists, NewLists),
	update_mutable(NewLists, OtherListsM),
	arg(1, Dom, Set),
	arg(1, OtherDom, OtherSet),
	(   Set==OtherSet -> Goals = []
	;   '$fd_dom_union'(Set, OtherSet, Union),
	    '$fd_dom_intersection'(Set, OtherSet, Intersection),
	    Intersection \== [],
	    set_to_range(Union, DomTerm),
	    update_mutable(DomTerm, OtherDomM),
	    Goals = [prune_and_propagate(Other,Intersection)]
	).

merge_lists('$fdlists'(0, _, _, _, _,  _, _), Lists2, Lists) :- !, Lists = Lists2.
merge_lists(Lists1, '$fdlists'(0, _, _, _, _,  _, _), Lists) :- !, Lists = Lists1.
merge_lists('$fdlists'(_, V,A1,A2,A4,A24,A8),
	    '$fdlists'(_, W,B1,B2,B4,B24,B8),
	    '$fdlists'(K,VW,C1,C2,C4,C24,C8)) :-
	VW is V\/W,
	not_entailed(A1, 0, K0,	 C1,  C1b),
	not_entailed(B1, K0, K1, C1b, []),
	not_entailed(A2, K1, K2, C2,  C2b),
	not_entailed(B2, K2, K3, C2b, []),
	not_entailed(A4, K3, K4, C4,  C4b),
	not_entailed(B4, K4, K5, C4b, []),
	not_entailed(A24,K5, K6, C24, C24b),
	not_entailed(B24,K6, K7, C24b,[]),
	not_entailed_val(A8, K7, K8,  C8,  C8b),
	not_entailed_val(B8, K8, K,   C8b, []).

not_entailed([], I, J) --> {J = I}.
not_entailed([X|Xs], I, K) -->
	({X = ix(_,_,_,Ent,_), nonvar(Ent)} -> {J = I}; [X], {J is I+1}),
	not_entailed(Xs, J, K).

not_entailed_val([], I, J) --> {J = I}.
not_entailed_val([X|Xs], I, K) -->
	({X = _-ix(_,_,_,Ent,_), nonvar(Ent)} -> {J = I}; [X], {J is I+1}),
	not_entailed_val(Xs, J, K).
*/

%%% Support for project_attributes and B&B search.

purify_vars([]).
purify_vars([X|Xs]) :-
	var(X), !,
	put_atts(X, -fd_attribute(_,_,_)),
	purify_vars(Xs).
purify_vars([_|Xs]) :-
	purify_vars(Xs).

%%% Support for global (specialized) constraints.

% Susp is a list of F(Var) terms, each representing that the constraint should be
% suspended on the variable Var.  F denotes the circumstances under which the constraint
% should be resumed:
% dom - resume when dom(Var) changes
% min - resume when min(Var) changes
% max - resume when max(Var) changes
% minmax - resume when min(Var) or max(Var) changes
% val - resume when Var becomes nonvar
fd_global(ModConstraint, State, Susp) :-
	prolog:get_module(ModConstraint, Constraint, Module),
	fd_global(Constraint, State, Susp, _, Module).

fd_global(Constraint, State, Susp, Global, Module) :-
	'$fd_global_created',
	dispatch_prune_and_enqueue(Constraint, State, NewState,
				   Ent, Q0, Q, GQ, GQTail1),
	Global = global(StateM,Constraint,StatusM,Ent,Module),
	create_mutable(2, StatusM),
	'$fd_global_told'(Global),
	create_mutable(NewState, StateM),
	(   var(Ent)
	->  '$fd_link'(Susp, Global)		% suspend initially
	;   true
	),
	evaluate(Q0, Q, GQ, GQTail1).

dispatch_prune_and_enqueue(Constraint, State, NewState,
			   Ent, Q0, Q, GQ0, GQ) :-
	dispatch_global_fast(Constraint, State, NewState, Actions),
	'$fd_prune_and_enqueue'(Actions, Ent, Q0, Q1, GQ0, GQ1, Calls),
	call_actions(Calls, Q1, Q, GQ1, GQ).

call_actions([], Q, Q, GQ, GQ).
call_actions([Goal|Goals], Q0, Q, GQ0, GQ) :-
	call_action(Goal, Q0, Q1, GQ0, GQ1),
	call_actions(Goals, Q1, Q, GQ1, GQ).

call_action(call(Goal), Q, Q, GQ, GQ) :- Goal.
call_action(X in R, Q0, Q, GQ0, GQ) :- 
	set_expression(R, S),
	'$fd_in_set'(X, S, Q0, Q, GQ0, GQ).
call_action(X in_set S, Q0, Q, GQ0, GQ) :- 
	'$fd_in_set'(X, S, Q0, Q, GQ0, GQ).
call_action(X=I, Q0, Q, GQ0, GQ) :-
	'$fd_range'(I, I, Set),
	'$fd_in_set'(X, Set, Q0, Q, GQ0, GQ).

/*
Moved to C since it must be atomic!

prune_and_queue([], _, Q, Q, GQ, GQ).
prune_and_queue([Action|Vs], Ent, Q0, Q, GQ0, GQ) :-
	prune_action(Action, Ent, Q0, Q1, GQ0, GQ1),
	prune_and_queue(Vs, Ent, Q1, Q, GQ1, GQ).

prune_action(exit, 1, Q, Q, GQ, GQ).
prune_action(fail, 0, Q, Q, GQ, GQ).
prune_action(call(G), _, Q, Q, GQ, GQ) :- G.
prune_action(X in R, _, Q0, Q, GQ0, GQ) :-
	set_expression(R, S),
	'$fd_in_set'(X, S, Q0, Q, GQ0, GQ).
prune_action(X in_set S, _, Q0, Q, GQ0, GQ) :-
	'$fd_in_set'(X, S, Q0, Q, GQ0, GQ).
prune_action(X=V, _, Q0, Q, GQ0, GQ) :-
	'$fd_in_interval'(X, V, V, Q0, Q, GQ0, GQ).
% new additions
prune_action(dom(X), _, Q0, Q, GQ0, GQ) :-
	'$fd_told'(X, 1, Q0, Q, GQ0, GQ).
prune_action(min(X), _, Q0, Q, GQ0, GQ) :-
	'$fd_told'(X, 3, Q0, Q, GQ0, GQ).
prune_action(max(X), _, Q0, Q, GQ0, GQ) :-
	'$fd_told'(X, 5, Q0, Q, GQ0, GQ).
prune_action(minmax(X), _, Q0, Q, GQ0, GQ) :-
	'$fd_told'(X, 7, Q0, Q, GQ0, GQ).
*/

% clpfd:dispatch_global(+Constraint, +State0, -State, -Actions)
% calls a user-defined constraint solver for a particular kind of constraint.
% Constraint is the original constraint;
% State0 is a term representing aux. info about this constraint at the time of its
% latest invocation.
% State represents updated aux. info;
% Actions is a list of terms of the following form:
%	exit - the constraint has been found entailed
%	fail - the constraint has been found disentailed
%	call(G) - call the Prolog goal G
%	X = R   - call(X in R)
%	X in R   - call(X in R)
%	X in_set R  - call(X in_set R)

dispatch_global_fast(in_set_iff(_,_,_), State, NewState, Actions) :- !,
	NewState = State,
	'$fd_in_set_iff'(State, Actions).
dispatch_global_fast('x*x=y'(X,Y), State, NewState, Actions) :- !,
	'x*x=y solver'(X, Y, State, NewState, Actions).
dispatch_global_fast('x*y=z'(X,Y,Z), State, NewState, Actions) :- !,
	'x*y=z solver'(X, Y, Z, State, NewState, Actions).
dispatch_global_fast('x/y=z'(X,Y,Z), State, NewState, Actions) :- !,
	'x/y=z solver'(X, Y, Z, State, NewState, Actions).
dispatch_global_fast('x mod y=z'(X,Y,Z), State, NewState, Actions) :- !,
	'x mod y=z solver'(X, Y, Z, State, NewState, Actions).
dispatch_global_fast(count(I,_,Rel,S), State, NewState, Actions) :- !,
	'count solver'(I, State, NewState, Rel, S, Actions).
dispatch_global_fast(scalar_product(_,_,Rel,_), State, NewState, Actions) :- !,
	'scalar_product solver'(Rel, State, NewState, Actions).
dispatch_global_fast(all_different(_), State, NewState, Actions) :- !,
	'$fd_all_different'(State, NewState, Actions).
dispatch_global_fast(all_distinct(_), State, NewState, Actions) :- !,
	'$fd_all_distinct'(State, NewState, Actions).
dispatch_global_fast(circuit(_,_), State, NewState, Actions) :- !,
	'$fd_circuit'(State, NewState, Actions).
dispatch_global_fast(element(_,_,_), State, NewState, Actions) :- !,
	'$fd_element'(State, NewState, Actions).
dispatch_global_fast(relation(_,_,_), State, NewState, Actions) :- !,
	'$fd_relation'(State, NewState, Actions).
dispatch_global_fast(assignment(_,_), State, NewState, Actions) :- !,
	'$fd_assignment'(State, NewState, Actions).
dispatch_global_fast(bool(F,_,_,_), State, NewState, Actions) :- !,
	'bool solver'(F, State, NewState, Actions).
dispatch_global_fast(task_order(_,_,Limit,EdgeF), State0, State, Actions) :- !,
	'task_order solver'(State0, State, Limit, EdgeF, Actions).
dispatch_global_fast(disjoint1(_,_), State, NewState, Actions) :- !,
	'disjoint1 solver'(State, NewState, Actions).
dispatch_global_fast(disjoint2(_,_), State, NewState, Actions) :- !,
	'disjoint2 solver'(State, NewState, Actions).
dispatch_global_fast(Global, State, NewState, Actions) :-
	dispatch_global(Global, State, NewState, Actions).

:- multifile
	dispatch_global/4.

:- dynamic
	dispatch_global/4.

/* fd_closure/2 */

% Given a list Vars of domain variables, Closure is the set
% of variables (including Vars) that can be transitively reached
% via constraints.
fd_closure(Vars, Closure) :-
	sort(Vars, Open),
	fd_closure(Open, Open, Closure).

/*** worst case O(N^2) version

fd_closure([], Closure, Closure).
fd_closure([X|Open], Closed, Closure) :-
	fd_neighbors(X, Neighs),
	ord_union(Closed, Neighs, Closed1, New),
	ord_union(Open, New, Open1),
	fd_closure(Open1, Closed1, Closure).

fd_neighbors(X, Neighs) :-
	var(X),
	clpfd:get_atts(X, fd_attribute(_,_,ListM)), !,
	get_mutable(Lists, ListM),
	Lists = '$fdlists'(_,_,L1,L2,L3,L4,L5),
	fd_neighbors(L1, S0, S1),
	fd_neighbors(L2, S1, S2),
	fd_neighbors(L3, S2, S3),
	fd_neighbors(L4, S3, S4),
	fd_neighbors(L5, S4, []),
	sort(S0, Neighs).
fd_neighbors(_, []).

fd_neighbors([]) --> [].
fd_neighbors([Item|L]) --> fd_neighbors(Item, L).

fd_neighbors(_-Item, L) --> !, fd_neighbors(Item, L).
fd_neighbors(iff(Item,_,_,_), L) --> !, fd_neighbors(Item, L).
fd_neighbors(ix(_,C,_,Ent,B,_,_), L) --> {var(Ent)}, !,
	{prolog:'$term_variables'(C, Vars)},
	list(Vars),
	(   {clpfd:'$get_attributes'(B, _, 1)} -> [B]
	;   []
	),
	fd_neighbors(L).
fd_neighbors(global(_,C,_,Ent,_), L) --> {var(Ent)}, !,
	{prolog:'$term_variables'(C, Vars)},
	list(Vars),
	fd_neighbors(L).
fd_neighbors(_, L) -->
	fd_neighbors(L).

list([]) --> [].
list([X|Xs]) --> [X], list(Xs).

***/

fd_closure([], Closure, Closure).
fd_closure([X|Open], Closed, Closure) :-
	fd_list_constraints([X|Open], Cs, []),
	prolog:term_variables(Cs, Neighs),
	ord_union(Closed, Neighs, Closed1, Open1),
	fd_closure(Open1, Closed1, Closure).

fd_neighbors(X, Neighs) :-
	fd_list_constraints([X], Cs, []),
	prolog:term_variables(Cs, Neighs).

fd_list_constraints([]) --> [].
fd_list_constraints([X|Xs]) -->
	{var(X),
	 get_atts(X, fd_attribute(_,_,ListM)), !,
	 get_mutable(Lists, ListM),
	 Lists = '$fdlists'(_,_,L1,L2,L3,L4,L5)},
	fd_neighbors(L1),
	fd_neighbors(L2),
	fd_neighbors(L3),
	fd_neighbors(L4),
	fd_neighbors(L5),
	fd_list_constraints(Xs).
fd_list_constraints([_|Xs]) -->
	fd_list_constraints(Xs).

%%% similar to suspensions/3, but we cannot assume that we
%%% have all zero-one variables already

fd_neighbors([]) --> [].
fd_neighbors([Item|L]) --> 
	fd_neighbors(Item, L).

fd_neighbors(_-Item, L) --> !, 
	fd_neighbors(Item, L).
fd_neighbors(iff(Item,_,_,_), L) --> !, 
	fd_neighbors(Item, L).
fd_neighbors(ix(_,C,_,Ent,B,_,_), L) --> {var(Ent)}, !,
	[B-C],
	fd_neighbors(L).
fd_neighbors(global(_,C,_,Ent,_), L) --> {var(Ent)}, !,
	[C],
	fd_neighbors(L).
fd_neighbors(_, L) -->
	fd_neighbors(L).

/* fd_copy_term/3 */

% Given a term Term containing domain variables, Template is a
% copy of the same term with all variables renamed to new variables
% such that executing Body will post constraints equivalent to those that
% Term is attached to.
% 
% For example:
% | ?- X in 0..1, Y in 10..11, X+5 #=< Y, clpfd:fd_copy_term(f(X,Y), Template, Body). 
% 
% Body = _A in_set[[0|1]],_B in_set[[10|11]],clpfd:'t>=u+c'(_B,_A,5),
% Template = f(_A,_B),
% X in 0..1,
% Y in 10..11 ? 
fd_copy_term(Term, Template, Body) :-
	prolog:'$term_variables'(Term, Vars0),
	fd_closure(Vars0, Vars0, Vars),
	all_suspensions(Vars, Goals, Goals2, Goals0, []),
	sort(Goals0, Goals2),
	andify(Goals, Body0),
	findall(Term-Body0, purify_vars(Vars), [Template-Body]).

all_suspensions([], S, S) --> [].
all_suspensions([X|Xs], S0, S) -->
	{get_atts(X, fd_attribute(_,DomM,ListsM))}, !,
	{get_mutable(Dom, DomM)},
	{get_mutable(Lists, ListsM)},
	{Dom = dom(Set,Min,Max,_)},
	{   Min==inf, Max==sup -> S0 = S1
	;   S0 = [X in_set Set | S1]
	},
	suspensions(Lists),
	all_suspensions(Xs, S1, S).
all_suspensions([_|Xs], S0, S) -->
	all_suspensions(Xs, S0, S).

